from bioscience.base import *

import math
import sys
import os
import threading
import warnings
import numpy as np
import time

def mcc(dataset, deviceCount = 0, mode = 1, debug = False):
    """
    Application of the Matthews Correlation Coefficient.
    
    :param dataset: The dataset object to be binarized.
    :type dataset: :class:`bioscience.base.models.Dataset`
    
    param deviceCount: Number of GPU devices to execute
    :type deviceCount: int
    
    :param mode: Type of execution of the algorithm: `mode=1` for sequential execution, `mode=2` for parallel execution on CPUs and `mode=3` for execution on a multi-GPU architecture.
    :type mode: int
    
    :return: A CorrelationModel object that stores values generated by a correlation method.
    :rtype: :class:`bioscience.base.models.CorrelationModel`      
    """

    oModel = None
    if (dataset is not None):
        sMode = ""
        if mode == 2: # NUMBA: CPU Parallel mode
            # To be developed
            sMode = "NUMBA - CPU Parallel mode (to be developed)"
        elif mode == 3: # NUMBA: GPU Parallel mode
            # To be developed
            sMode = "NUMBA - GPU Parallel mode (to be developed)"
        else: # Sequential mode
            oModel = __mccSequential(dataset, debug)
            deviceCount = 0
            sMode = "CPU Sequential"
    
    return oModel

def __mccSequential(dataset, debug):
    iRows = dataset.data.shape[0]
    iCols = dataset.data.shape[1]
    fExecutionTime = None
    
    maxPairs = 0        
    for i in range(iRows):
        for j in range(i + 1, iRows):
            maxPairs += 1
    
    resultsCorrelation = np.zeros(maxPairs)    
    
    if debug == True:
        start_time = time.time()
        
    for pattern, value in enumerate(resultsCorrelation):
        
        # Get R1 and R2 from index results vector
        r1 = 0
        r2 = -1
        auxPat = pattern - iRows + 1
        
        if auxPat < 0:
            r2 = auxPat + iRows

        j = iRows - 2
        while r2 == -1:
            auxPat -= j
            r1 += 1
            if auxPat < 0:
                r2 = (j + auxPat) + (r1 + 1)
            j -= 1
        
        if r1 < iRows and r2 < iRows:
            
            # Confusion matrix
            TP = np.sum((dataset.data[r1] == 1) & (dataset.data[r2] == 1))
            TN = np.sum((dataset.data[r1] == 0) & (dataset.data[r2] == 0))
            FP = np.sum((dataset.data[r1] == 0) & (dataset.data[r2] == 1))
            FN = np.sum((dataset.data[r1] == 1) & (dataset.data[r2] == 0))

            # Calculate MCC value
            num = (TP * TN) - (FP * FN)
            den = np.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN))
            
            mccValue = None
            if den != 0:
                mccValue = num / den
        
            resultsCorrelation[pattern] = mccValue
    
    if debug == True:
        end_time = time.time()
        fExecutionTime = end_time - start_time
    
    oCorrelationResults = CorrelationModel(name=MCC, results=resultsCorrelation, rows = iRows, executionTime=fExecutionTime)
    return oCorrelationResults