from bioscience.base import *

import math
import sys
import os
import threading
import warnings
import numpy as np
import time

def pbc(dataset, deviceCount = 0, mode = 1, debug = False):
    """
    Application of the Point-biserial correlation (PBC) method.
    
    :param dataset: The dataset object to be binarized.
    :type dataset: :class:`bioscience.base.models.Dataset`
    
    param deviceCount: Number of GPU devices to execute
    :type deviceCount: int
    
    :param mode: Type of execution of the algorithm: `mode=1` for sequential execution, `mode=2` for parallel execution on CPUs and `mode=3` for execution on a multi-GPU architecture.
    :type mode: int
    
    :return: A CorrelationModel object that stores values generated by a correlation method.
    :rtype: :class:`bioscience.base.models.CorrelationModel`      
    """

    oModel = None
    if (dataset is not None):
        sMode = ""
        if mode == 2: # NUMBA: CPU Parallel mode
            # To be developed
            sMode = "NUMBA - CPU Parallel mode (to be developed)"
        elif mode == 3: # NUMBA: GPU Parallel mode
            # To be developed
            sMode = "NUMBA - GPU Parallel mode (to be developed)"
        else: # Sequential mode
            oModel = __pbcSequential(dataset, debug)
            deviceCount = 0
            sMode = "CPU Sequential"
    
    return oModel

def __pbcSequential(dataset, debug):
    iRows = dataset.data.shape[0]
    iCols = dataset.data.shape[1]
    fExecutionTime = None
    
    maxPairs = 0        
    for i in range(iRows):
        for j in range(i + 1, iRows):
            maxPairs += 1
    
    resultsCorrelation = np.zeros(maxPairs)    
    
    if debug == True:
        start_time = time.time()
        
    for pattern, value in enumerate(resultsCorrelation):
        
        # Get R1 and R2 from index results vector
        r1 = 0
        r2 = -1
        auxPat = pattern - iRows + 1
        
        if auxPat < 0:
            r2 = auxPat + iRows

        j = iRows - 2
        while r2 == -1:
            auxPat -= j
            r1 += 1
            if auxPat < 0:
                r2 = (j + auxPat) + (r1 + 1)
            j -= 1
        
        if r1 < iRows and r2 < iRows:
            
            
            elementsR2_0 = []
            elementsR2_1 = []
            
            addValR2_0 = 0
            addValR2_1 = 0
            
            for i in range(iCols):
                if dataset.data[r2][i] == 0:
                    elementsR2_0.append(dataset.data[r1][i])
                    addValR2_0 += dataset.data[r1][i]
                elif dataset.data[r2][i] == 1:
                    elementsR2_1.append(dataset.data[r1][i])
                    addValR2_1 += dataset.data[r1][i]
            
            mean_R2_0 = addValR2_0 / len(elementsR2_0) if elementsR2_0 else 0
            mean_R2_1 = addValR2_1 / len(elementsR2_1) if elementsR2_1 else 0
            diffMean = mean_R2_1 - mean_R2_0
            
            n0 = len(elementsR2_0)
            n1 = len(elementsR2_1)
            side = math.sqrt((n1*n0) / math.pow(iCols,2))
            
            meanR1 = 0
            for i in range(iCols):
                meanR1 += dataset.data[r1][i]
            meanR1 /= iCols
            
            sum = 0
            for i in range(iCols):
                a = dataset.data[r1][j] - meanR1
                sum += math.pow(a, 2)
            bottomSide = math.sqrt(sum / iCols)
            
            pbcValue = None
            if bottomSide != 0:
                pbcValue = (diffMean / bottomSide) * side
            
            resultsCorrelation[pattern] = pbcValue
    
    if debug == True:
        end_time = time.time()
        fExecutionTime = end_time - start_time
    
    oCorrelationResults = CorrelationModel(name=PBC, results=resultsCorrelation, rows = iRows, executionTime=fExecutionTime)
    return oCorrelationResults